@@ -1,8 +1,8 @@
-#define VERSION "1.81"
+#define VERSION "1.84a"
 
 /*
  *  units, a program for units conversion
- *  Copyright (C) 1996, 1997, 1999, 2000, 2001, 2002 
+ *  Copyright (C) 1996, 1997, 1999, 2000, 2001, 2002, 2003, 2004 
  *  Free Software Foundation, Inc
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -50,17 +50,24 @@
 #define UNITMATCH "?"           /* Command to request conformable units */
 #define DEFAULTPAGER "more"     /* Default pager program */
 #define DEFAULTLOCALE "en_US"   /* Default locale */
+#define MAXINCLUDE 5            /* Max depth of include files */
+#define MAXFILES 25             /* Max number of units files on command line */
+#define NODIM "!dimensionless"  /* Marks dimensionless primitive units, such */
+				/* as the radian, which are ignored when */
+                                /* doing unit comparisons. */
 
 char *pager;                    /* Actual pager (from PAGER environment var) */
 char *mylocale;                 /* Locale in effect (from LOCALE env var) */
 char *numformat = "%.8g";	/* printf format for numeric output */
 char *powerstring = "^";	/* Exponent character used in output */
 int quiet = 0;                  /* Flag for supressing prompting (-q option) */
+int terseout = 0;               /* Flag for terse output (-t option) */
 int unitcheck = 0;              /* Flag for unit checking, set to 1 for 
                                    regular check, set to 2 for verbose check */
 int verbose = 0;                /* Flag for verbose operation */
 int strictconvert = 0;          /* Strict conversion (disables reciprocals) */
-char *userfile = NULL;          /* User specified units file name */
+int minusminus = 1;             /* Should '-' character give subtraction */
+char *unitsfiles[MAXFILES+1];   /* Null terminated list of units file names */
 char *progname="units";         /* Used in error messages */
 char *queryhave = "You have: "; /* Prompt text for units to convert from */
 char *querywant = "You want: "; /* Prompt text for units to convert to */
@@ -77,7 +84,7 @@ char *errormsg[]={"Successful completion",
                   "Parse error",           
                   "Product overflow",      
                   "Unit reduction error (bad unit definition)",
-                  "Illegal sum of non-conformable units",
+                  "Illegal sum or difference of non-conformable units",
                   "Unit not dimensionless",
 		  "Unit not a root",
 		  "Unknown unit",
@@ -88,7 +95,10 @@ char *errormsg[]={"Successful completion",
 		  "Nonlinear unit definition has unit error",
 		  "No inverse defined",
 		  "Parser memory overflow (recursive function definition?)",
-		  "Argument wrong dimension or bad nonlinear unit definition"
+		  "Argument wrong dimension or bad nonlinear unit definition",
+                  "Unable to open units file",
+		  "Units file contains errors",
+		  "Memory allocation error"
                   };
 
 char *irreducible=0;            /* Name of last irreducible unit */
@@ -100,6 +110,7 @@ struct unitlist {
    char *name;			/* unit name */
    char *value;			/* unit value */
    int linenumber;              /* line in units data file where defined */
+   char *file;                  /* file where defined */ 
    struct unitlist *next;	/* next item in list */
 } *utab[HASHSIZE];
 
@@ -111,6 +122,7 @@ struct prefixlist {
    char *name;			/* prefix name */
    char *value;			/* prefix value */
    int linenumber;              /* line in units data file where defined */
+   char *file;                  /* file where defined */ 
    struct prefixlist *last;	/* last item in list--only set in first item */
    struct prefixlist *next;   	/* next item in list */
 } *ptab[PREFIXTABSIZE];
@@ -121,7 +133,6 @@ struct prefixlist {
 struct func *firstfunc = 0;    /* Base of linked list of functions */
 struct func *lastfunc = 0;     /* Last entry in linked list of functions */
 
-
 /* 
    Used for passing parameters to the parser when we are in the process
    of parsing a unit function.  If function_parameter is non-nil, then 
@@ -338,99 +349,138 @@ isfunction(char *str)
 /* Print out error message encountered while reading the units file. */
 
 void
-readerror(int linenum, const char *filename)
+readerror(FILE *errfile, int linenum, const char *filename)
 {
-   fprintf(stderr, "%s: error in units file '%s' line %d\n",
-	   progname, filename, linenum);
+  if (errfile)
+    fprintf(errfile, "%s: error in units file '%s' line %d\n",
+	    progname, filename, linenum);
 }
 
 
-/* Read in units data.  
+/* 
+   Read in units data.  
 
-   If unitcheck is set then primitive units are set to "1" so all reductions
-   should give a plain number.  */
+   file - Filename to load
+   errfile - File to receive messages about errors in the units database.  
+             Set it to 0 to suppress errors. 
+   unitcount, prefixcount, funccount - Return statistics to the caller.  
+                                       Must initialize to zero before calling.
+   depth - Used to prevent recursive includes.  Call with it set to zero.
 
-void
-readunits() /* char *userfile) */
+
+   The global variable progname is used in error messages.  
+*/
+
+int
+readunits(char *file, FILE *errfile, 
+          int *unitcount, int *prefixcount, int *funccount, int depth)
 {
    struct prefixlist *pfxptr;
    struct unitlist *uptr;
    FILE *unitfile;
-   char *line, *lineptr, *unitsfilename, *unitdef, *unitname;
-   int len, linenum, linebufsize;
+   char *line, *lineptr, *unitdef, *unitname, *permfile;
+   int len, linenum, linebufsize, goterr;
    unsigned hashval, pval;
-   int unitcount, prefixcount, funccount;
+   int locunitcount, locprefixcount, locfunccount;
    struct func *funcentry;
    int wronglocale = 0;   /* If set then we are currently reading data */
-                          /* for the wrong locale so we should skip it */
-   unitcount = 0;
-   prefixcount = 0;
-   funccount  = 0;
+   int inlocale = 0;      /* for the wrong locale so we should skip it */
+   locunitcount = 0;
+   locprefixcount = 0;
+   locfunccount  = 0;
    linenum = 0;
    linebufsize = 0;
+   goterr = 0;
 
    growbuffer(&line,&linebufsize);
 
-   if (userfile) {
-      unitfile = fopen(userfile, "rt");
-      if (!unitfile) {
-	 fprintf(stderr, "%s: unable to open units file '%s'.  ",
-		 progname, userfile);
-         perror(0);
-	 exit(1);
-      }
-      unitsfilename = userfile;
-   } else {
-      unitfile = fopen(UNITSFILE, "rt");
-      unitsfilename = UNITSFILE;
-      if (!unitfile) {
-	 char *direc, *env;
-	 char *filename;
-	 char separator[2];
-
-	 env = getenv("PATH");
-         filename = mymalloc(strlen(env)+strlen(UNITSFILE)+2,"(readunits)");
-	 if (env) {
-	    if (strchr(env, ';'))	/* MS-DOS */
-	       strcpy(separator, ";");
-	    else		        /* UNIX */
-	       strcpy(separator, ":");
-	    direc = strtok(env, separator);
-	    while (direc) {
-	       strcpy(filename, "");
-	       strcat(filename, direc);
-	       strcat(filename, "/");
-	       strcat(filename, UNITSFILE);
-	       unitfile = fopen(filename, "rt");
-	       if (unitfile){
-                  unitsfilename = dupstr(filename);
-		  userfile = unitsfilename;
-		  break;
-	       }
-	       direc = strtok(NULL, separator);
-	    }
-	 }
-	 if (!unitfile) {
-	    fprintf(stderr, "%s: can't find units file '%s'\n",
-		    progname, UNITSFILE);
-	    exit(1);
-	 }
-         free(filename);
-      }
-   }
+   permfile = dupstr(file);                 /* This is a permanent copy to 
+                                               reference in the database.
+                                               It is never freed. */
+   unitfile = fopen(file, "rt");
+   if (!unitfile) 
+     return E_FILE;
    while (!feof(unitfile)) {
       if (!fgetslong(&line, &linebufsize, unitfile, &linenum)) 
         break;
       if (*line == COMMANDCHAR) {         /* Process units.dat commands    */
         unitname = strtok(line+1, WHITE);
-	if (!strcmp(unitname,"locale")){  /* The only commands right now   */
-	  unitname = strtok(0, WHITE);    /* provide locale support.       */
-	  if (strcmp(unitname,mylocale))  /* locales don't match           */
-	    wronglocale = 1;
+	if (!strcmp(unitname,"locale")){ 
+	  unitname = strtok(0, WHITE); 
+	  if (!*unitname) {
+	    if (errfile)
+	      fprintf(errfile, 
+		      "%s: no locale specified on line %d of '%s'\n",
+		      progname, linenum, file);
+	    goterr=1;
+	  } else if (inlocale){
+	    if (errfile)
+	      fprintf(errfile,
+		      "%s: nested locales not allowed, line %d of '%s'\n",
+		      progname, linenum, file);
+	    goterr=1;
+	  } else {
+	    inlocale = 1;
+	    if (strcmp(unitname,mylocale))  /* locales don't match           */
+	      wronglocale = 1;
+	  }
+	  continue;
 	} 
-	else if (!strcmp(unitname, "endlocale"))
+	else if (!strcmp(unitname, "endlocale")){
+	  if (!inlocale){
+	    if (errfile)
+	      fprintf(errfile, 
+		      "%s: unmatched !endlocale on line %d of '%s'\n",
+		      progname, linenum, file);
+	    goterr=1;
+	  }
 	  wronglocale = 0;
-	else readerror(linenum,unitsfilename);
+	  inlocale = 0;
+	  continue;
+	}
+	if (wronglocale)
+	  continue;
+        if (!strcmp(unitname, "include")){
+          if (depth>MAXINCLUDE){
+	    if (errfile)
+	      fprintf(errfile, 
+		      "%s: max include depth of %d exceeded in file '%s' line %d\n",
+		      progname, MAXINCLUDE, file, linenum);
+	    goterr=1;
+	  } else {
+	    int readerr;
+	    char *includefile;
+	    unitname = strtok(0, WHITE); 
+	    includefile = mymalloc(strlen(file)+strlen(unitname)+1, "(readunits)");
+            if (strchr(unitname, '/') || strchr(unitname, '\\'))
+	      strcpy(includefile,unitname);
+	    else {
+              char *pathend;
+	      strcpy(includefile, file);
+	      pathend = includefile + strlen(includefile) - 1;
+	      while(pathend != includefile && !strchr("/\\", *pathend))
+		pathend--;
+	      if (strchr("/\\", *pathend))
+		pathend++;
+	      strcpy(pathend, unitname);
+	    }
+	    readerr = readunits(includefile, errfile, unitcount, prefixcount, 
+				funccount, depth+1);
+	    if (readerr == E_MEMORY) 
+	      return readerr;
+	    if (readerr == E_FILE) {
+	      if (errfile)
+		fprintf(errfile, "%s: unable to open included file '%s' at line %d of file '%s\n", progname, includefile, linenum, file);
+	    }
+	    
+	    if (readerr)
+	      goterr = 1;
+	    free(includefile);
+	  }
+	} else {                             /* not a valid command */
+	  readerror(errfile,linenum,file);
+	  goterr=1;
+	}
 	continue;
       }	
       if (wronglocale)
@@ -442,26 +492,39 @@ readunits() /* char *userfile) */
 	continue;
       unitdef = strtok(NULL, "\n");
       if (!unitdef){
-        readerror(linenum,unitsfilename);
+        readerror(errfile,linenum,file);
+	goterr=1;
         continue;
       }
       unitdef = removepadding(unitdef);
       if (!*unitdef){ 
-        readerror(linenum,unitsfilename);
+        readerror(errfile,linenum,file);
+	goterr=1;
         continue;
       }
       
       len = strlen(unitname);      
+
       if (unitname[len - 1] == '-') {	/* it's a prefix definition */
          unitname[len - 1] = 0;
+	 if (strchr("0123456789.", unitname[0])){
+	   if (errfile)
+	     fprintf(errfile,
+	     "%s: unit '%s' on line %d of '%s' ignored.  It starts with a digit\n", 
+	     progname, unitname, linenum, file);
+  	     goterr=1;
+	     continue;
+	 }
 	 if (pfxptr = plookup(unitname)) {  /* already there: redefinition */
-	    if (!strcmp(pfxptr->name, unitname))
-	       fprintf(stderr,
-   	       "%s: redefinition of prefix '%s-' on line %d ignored.\n",
-		       progname, unitname, linenum);
-	    else
-	       fprintf(stderr, "%s: prefix '%s-' on line %d is hidden by earlier definition of '%s-'.\n",
-		       progname, unitname, linenum, pfxptr->name);
+ 	    goterr=1;
+            if (errfile)
+	      if (!strcmp(pfxptr->name, unitname))
+		fprintf(errfile,
+   	        "%s: redefinition of prefix '%s-' on line %d of '%s' ignored.\n",
+		       progname, unitname, linenum, file);
+	      else
+	         fprintf(errfile, "%s: prefix '%s-' on line %d of '%s' is hidden by earlier definition of '%s-'.\n",
+		       progname, unitname, linenum, file, pfxptr->name);
 	    continue;
 	 } 
 
@@ -472,6 +535,7 @@ readunits() /* char *userfile) */
 	 pfxptr->len = len - 1;
 	 pfxptr->value = dupstr(unitdef);
          pfxptr->linenumber = linenum;
+	 pfxptr->file = permfile;
 	 /*
 	    Install prefix name/len/value in list
 	    Order is FIFO, so a prefix that is a substring of another
@@ -487,20 +551,38 @@ readunits() /* char *userfile) */
 	 else
 	    ptab[pval]->last->next = pfxptr;
 	 ptab[pval]->last = pfxptr;
-	 prefixcount++;
+	 locprefixcount++;
       } else if (strchr(unitname,'[')){ /* table definition  */
 	char *start, *end;
 	int tablealloc, tabpt;
         struct pair *tab;
-	int goterr;
+	int tableerr;
 
-	goterr=0;
+	tableerr=0;
 	start  = strchr(unitname,'[');
 	end = strchr(unitname,']');
 	*start++=0;
+	if (strchr("0123456789.", unitname[0])){
+	  if (errfile)
+	    fprintf(errfile,
+            "%s: unit '%s' on line %d ignored.  It starts with a digit\n", 
+		     progname, unitname, linenum);
+	  goterr=1;
+	  continue;
+	}
 	if (!end || strlen(end)>1){
-	  fprintf(stderr,"%s: missing ']' in units file '%s' line %d\n",
-		 progname,unitsfilename,linenum);
+	  if (errfile)
+ 	    fprintf(errfile,"%s: missing ']' in units file '%s' line %d\n",
+		    progname,file,linenum);
+	  goterr=1;
+	  continue;
+	} 
+        if (fnlookup(unitname,strlen(unitname))){
+	  if (errfile)
+	    fprintf(errfile,
+		  "%s: redefinition of unit '%s' on line %d of file '%s' ignored\n",
+		  progname, unitname, linenum, file);
+	  goterr=1;
 	  continue;
 	}
 	*end=0;
@@ -516,42 +598,47 @@ readunits() /* char *userfile) */
 	    tablealloc+=20;
 	    tab = (struct pair *)realloc(tab,sizeof(struct pair)*tablealloc);
 	    if (!tab){
-	      fprintf(stderr, "%s: memory allocation error (readunits)\n",
-		      progname);  
-	      exit(3); 
+	      if (errfile)
+	        fprintf(errfile, "%s: memory allocation error (readunits)\n",
+		        progname);  
+	      return E_MEMORY;
 	    }
 	  }
 	  tab[tabpt].location = strtod(start,&end);
 	  if (start==end)
 	    break;
 	  if (tabpt>0 && tab[tabpt].location<=tab[tabpt-1].location){
-	    fprintf(stderr,"%s: points don't increase (%.8g to %.8g) in units file '%s' line %d\n",
+	    if (errfile)
+  	      fprintf(errfile,"%s: points don't increase (%.8g to %.8g) in units file '%s' line %d\n",
 		    progname, tab[tabpt-1].location, tab[tabpt].location,
-		    unitsfilename, linenum);
-	    goterr=1;
+		    file, linenum);
+	    tableerr=1;
 	    break;
 	  }
 	  start=end+strspn(end," \t");
 	  tab[tabpt].value = strtod(start,&end);
 	  if (start==end){
-	    fprintf(stderr,"%s: missing value after %.8g in units file '%s' line %d\n",
-		    progname, tab[tabpt].location, unitsfilename, linenum);
-	    goterr=1;
+	    if (errfile)
+	      fprintf(errfile,"%s: missing value after %.8g in units file '%s' line %d\n",
+		    progname, tab[tabpt].location, file, linenum);
+	    tableerr=1;
 	    break;
 	  }
 	  tabpt++;
 	  start=end+strspn(end," \t,");
 	}
-	if (goterr){
+	if (tableerr){
 	  free(tab);
 	  free(funcentry->name);
 	  free(funcentry->tableunit);
 	  free(funcentry);
+	  goterr=1;
 	} else {
 	  funcentry->tablelen = tabpt;
 	  funcentry->table = tab;
           funcentry->linenumber = linenum;
-	  funccount++;
+	  funcentry->file = permfile;
+	  locfunccount++;
 	  addfunction(funcentry);
 	}
       } else if (strchr(unitname,'(')){ /* function definition */
@@ -560,10 +647,30 @@ readunits() /* char *userfile) */
          start = strchr(unitname,'(');
          end = strchr(unitname,')');
          *start++ = 0;
+
+	 if (strchr("0123456789.", unitname[0])){
+	   if (errfile)
+	     fprintf(errfile,
+	     "%s: unit '%s' on line %d of '%s' ignored.  It starts with a digit\n", 
+		   progname, unitname, linenum, file);
+	   goterr=1;
+	   continue;
+	 }
+
          if (!end || strlen(end)>1){
-	   fprintf(stderr,
+	   if (errfile)
+	     fprintf(errfile,
 		   "%s: bad function definition of '%s' in '%s' line %d\n",
-		   progname,unitname,unitsfilename,linenum);
+		   progname,unitname,file,linenum);
+	   goterr=1;
+	   continue;
+	 }
+	 if (fnlookup(unitname,strlen(unitname))){
+	   if (errfile)
+	     fprintf(errfile,
+		   "%s: redefinition of unit '%s' on line %d of '%s' ignored\n",
+		   progname, unitname, linenum, file);
+	   goterr=1;
 	   continue;
 	 }
  	 funcentry = (struct func*)mymalloc(sizeof(struct func),"(readunits)");
@@ -578,9 +685,11 @@ readunits() /* char *userfile) */
 	     *inv++=0;
 	   if (!end || (inv && (end-inv<0))){
 	     free(funcentry);
-	     fprintf(stderr,
+	     if (errfile)
+	       fprintf(errfile,
    	         "%s: expecting ']' in definition of '%s' in '%s' line %d\n",
-		 progname, unitname, unitsfilename, linenum);
+		 progname, unitname, file, linenum);
+	     goterr=1;
 	     continue;
 	   }
 	   *end=0;
@@ -605,34 +714,41 @@ readunits() /* char *userfile) */
          } 
 	 else 
 	   funcentry->inverse.def = 0;
-         funccount++;
+         locfunccount++;
 	 funcentry->linenumber = linenum;
+	 funcentry->file = permfile;
 	 addfunction(funcentry);
       } else {	/* it is a unit definition */
 
   	 /* Units that end in [2-9] can never be accessed */
 
  	 if (strchr("23456789", unitname[strlen(unitname)-1])){
-	   fprintf(stderr, 
-           "%s: unit '%s' on line %d ignored.  It ends with a nonzero digit\n",
-           progname, unitname, linenum);
+	   if (errfile)
+	     fprintf(errfile,
+             "%s: unit '%s' on line %d of '%s' ignored.  It ends with a nonzero digit\n",
+             progname, unitname, linenum, file);
+   	   goterr=1;
 	   continue;
 	 }
 
 	 if (strchr("0123456789.", unitname[0])){
-	   fprintf(stderr,
-	   "%s: unit '%s' on line %d ignored.  It starts with a digit\n", 
-		   progname, unitname, linenum);
+	   if (errfile)
+	     fprintf(errfile,
+		     "%s: unit '%s' on line %d of '%s' ignored.  It starts with a digit\n", 
+		     progname, unitname, linenum, file);
+	   goterr=1;
 	   continue;
 	 }
 
          /* Is it a redefinition? */
 
 	 if (ulookup(unitname)) {
-	    fprintf(stderr,
-		    "%s: redefinition of unit '%s' on line %d ignored\n",
-		    progname, unitname, linenum);
-    	    continue;
+	   if (errfile)
+	     fprintf(errfile,
+		    "%s: redefinition of unit '%s' on line %d of '%s' ignored\n",
+		    progname, unitname, linenum, file);
+	   goterr=1;
+	   continue;
 	 }
 
          /* make new units table entry */
@@ -641,20 +757,27 @@ readunits() /* char *userfile) */
 	 uptr->name = dupstr(unitname);
 	 uptr->value = dupstr(unitdef);
          uptr->linenumber = linenum;
+	 uptr->file = permfile;
 
 	 /* install unit name/value pair in list */
 
 	 hashval = uhash(uptr->name);
 	 uptr->next = utab[hashval];
 	 utab[hashval] = uptr;
-	 unitcount++;
+	 locunitcount++;
       }
    }
    fclose(unitfile);
    free(line);
-   if (!quiet)
-     printf("%d units, %d prefixes, %d nonlinear units\n\n", unitcount, prefixcount,
-	    funccount);
+   if (unitcount)
+     *unitcount+=locunitcount;
+   if (prefixcount)
+     *prefixcount+=locprefixcount;
+   if (funccount)
+     *funccount+=locfunccount;
+   if (goterr)
+     return E_BADFILE;
+   else return 0;
 }
 
 /* Initialize a unit to be equal to 1. */
@@ -798,14 +921,22 @@ cancelunit(struct unittype *theunit)
    buffer created to hold the data.  This buffer grows as needed during 
    program execution.  
 
-   Don't pass the output of lookupunit() back into the function again,
-   however, because the buffer may be overwritten producing incorrect
-   results.
+   Note that if you pass the output of lookupunit() back into the function
+   again you will get correct results, but the data you passed in may get
+   clobbered if it happened to be the internal buffer.  
 */
 
 static int bufsize=0;
 static char *buffer;  /* buffer for lookupunit answers with prefixes */
 
+
+/* 
+  Plural rules for english: add -s
+  after x, sh, ch, ss   add -es
+  -y becomes -ies except after a vowel when you just add -s as usual
+*/
+  
+
 char *
 lookupunit(char *unit,int prefixok)
 {
@@ -838,6 +969,17 @@ lookupunit(char *unit,int prefixok)
             return buffer;
 	 }
       }
+      if (strlen(copy)>2 && copy[strlen(copy) - 1] == 'i') {
+	 copy[strlen(copy) - 1] = 'y';
+	 if (lookupunit(copy,prefixok)){
+	    while (strlen(copy)+1 > bufsize) {
+	       growbuffer(&buffer,&bufsize);
+	    }
+            strcpy(buffer,copy);
+	    free(copy);
+            return buffer;
+	 }
+      }
       free(copy);
    }
    if (prefixok && (pfxptr = plookup(unit))) {
@@ -1043,21 +1185,65 @@ reduceunit(struct unittype *theunit)
    return 0;
 }
 
-/* Compare two product lists, return zero if they match and one if
-   they do not match.  They may contain embedded NULLUNITs which are
-   ignored in the comparison. */
+/* 
+   Returns one if the argument unit is defined in the data file as a   
+  dimensionless unit.  This is determined by comparing its definition to
+  the string NODIM.  
+*/
 
 int
-compareproducts(char **one, char **two)
+ignore_dimless(char *name)
 {
+  struct unitlist *ul;
+  if (!name) 
+    return 0;
+  ul = ulookup(name);
+  if (ul && !strcmp(ul->value, NODIM))
+    return 1;
+  return 0;
+}
+
+int 
+ignore_nothing(char *name)
+{
+  return 0;
+}
+
+
+int
+ignore_primitive(char *name)
+{
+  struct unitlist *ul;
+  if (!name) 
+    return 0;
+  ul = ulookup(name);
+  if (ul && strchr(ul->value, PRIMITIVECHAR))
+    return 1;
+  return 0;
+}
+
+
+/* 
+   Compare two product lists, return zero if they match and one if
+   they do not match.  They may contain embedded NULLUNITs which are
+   ignored in the comparison. Units defined as NODIM are also ignored
+   in the comparison.
+*/
+
+int
+compareproducts(char **one, char **two, int (*isdimless)(char *name))
+{
+   int oneblank, twoblank;
    while (*one || *two) {
-      if (!*one && *two != NULLUNIT)
+      oneblank = (*one==NULLUNIT) || isdimless(*one);    
+      twoblank = (*two==NULLUNIT) || isdimless(*two);    
+      if (!*one && !twoblank)
 	 return 1;
-      if (!*two && *one != NULLUNIT)
+      if (!*two && !oneblank)
 	 return 1;
-      if (*one == NULLUNIT)
+      if (oneblank)
 	 one++;
-      else if (*two == NULLUNIT)
+      else if (twoblank)
 	 two++;
       else if (strcmp(*one, *two))
 	 return 1;
@@ -1072,11 +1258,12 @@ compareproducts(char **one, char **two)
    must be reduced, sorted and canceled for this to work.  */
 
 int
-compareunits(struct unittype *first, struct unittype *second)
+compareunits(struct unittype *first, struct unittype *second, 
+	     int (*isdimless)(char *name))
 {
    return
-      compareproducts(first->numerator, second->numerator) ||
-      compareproducts(first->denominator, second->denominator);
+      compareproducts(first->numerator, second->numerator, isdimless) ||
+      compareproducts(first->denominator, second->denominator, isdimless);
 }
 
 
@@ -1151,7 +1338,7 @@ unit2num(struct unittype *input)
   initializeunit(&one);
   if (err=completereduce(input))
     return err;
-  if (compareunits(input,&one))
+  if (compareunits(input,&one,ignore_nothing))
     return E_NOTANUMBER;
   freeunit(input);
   return 0;
@@ -1295,7 +1482,7 @@ addunit(struct unittype *unita, struct unittype *unitb)
     return err;
   if (err=completereduce(unitb))
     return err;
-  if (compareunits(unita,unitb))
+  if (compareunits(unita,unitb,ignore_nothing))
     return E_BADSUM;
   unita->factor += unitb->factor;
   freeunit(unitb);
@@ -1399,7 +1586,7 @@ evalfunc(struct unittype *theunit, struct func *infunc, int inverse)
        err = completereduce(&result);
        if (err)
 	 return E_BADTABLE;
-       if (compareunits(&result, theunit))
+       if (compareunits(&result, theunit, ignore_nothing))
 	 return E_BADFUNCARG;
      }
      save_value = parameter_value;
@@ -1509,11 +1696,10 @@ showfunc(char *havestr, struct unittype *have, struct func *fun)
        printf("Function evaluation error (bad function definition)\n");
      return 1;
    }
-  
    if (verbose)
      printf("\t%s = %s(", havestr, fun->inverse.param);
    else 
-     putchar('\t');
+     if (!terseout) putchar('\t');
    showunit(have);
    if (verbose)
      putchar(')');
@@ -1535,7 +1721,7 @@ showanswer(char *havestr,struct unittype *have,
    doingrec=0;
    havestr = removepadding(havestr);
    wantstr = removepadding(wantstr);
-   if (compareunits(have, want)) {
+   if (compareunits(have, want, ignore_dimless)) {
         char **src,**dest;
 
         invhave.factor=1/have->factor;
@@ -1545,7 +1731,7 @@ showanswer(char *havestr,struct unittype *have,
         for(src=have->denominator,dest=invhave.numerator;*src;src++,dest++)
            *dest=*src;
         *dest=0;
-        if (strictconvert || compareunits(&invhave, want)){
+        if (strictconvert || compareunits(&invhave, want, ignore_dimless)){
 	  printf("conformability error\n");
 	  if (verbose) printf("\t%s = ",havestr);
 	  else putchar('\t');
@@ -1576,13 +1762,15 @@ showanswer(char *havestr,struct unittype *have,
      }
    }   
    if (verbose) printf("\t%s%s%s = ",left,havestr,right);
-   else printf("\t* ");
+   else if (!terseout) printf("\t* ");
    printf(numformat, have->factor / want->factor);
-   if (verbose) 
-     printf("%s %s\n\t%s%s%s = (1 / ",sep,wantstr,left,havestr,right);
-   else printf("\n\t/ ");
-   printf(numformat, want->factor / have->factor);
-   if (verbose) printf(")%s %s",sep,wantstr);
+   if (!terseout){
+     if (verbose) 
+       printf("%s %s\n\t%s%s%s = (1 / ",sep,wantstr,left,havestr,right);
+     else printf("\n\t/ ");
+     printf(numformat, want->factor / have->factor);
+     if (verbose) printf(")%s %s",sep,wantstr);
+   }
    putchar('\n');
    return 0;
 }
@@ -1660,17 +1848,19 @@ checkfunc(struct func *infunc, int verbose)
 
 
 /* 
-   Cycle through all units and prefixes and attempt to reduce each one to 1.
-   Print a message for all units which do not reduce to 1.  
+   Check that all units and prefixes are reducible to primitive units and that
+   function definitions are valid and have correct inverses.  A message is
+   printed for every unit that does not reduce to primitive units.
 
-   (When this function is called, all of the primitive units will have been 
-   defined equal to 1, so all valid unit definitions should reduce to 1.)
+   This function has the side effect of defining all primitive units
+   in the database to have the value "1".
 */
 
+
 void 
 checkunits(int verbosecheck)
 {
-  struct unittype have,one;
+  struct unittype have,second,one;
   struct unitlist *uptr;
   struct prefixlist *pptr;
   struct func *funcptr;
@@ -1679,20 +1869,10 @@ checkunits(int verbosecheck)
   initializeunit(&one);
 
   /* Check all functions for valid definition and correct inverse */
-
+  
   for(funcptr=firstfunc;funcptr;funcptr=funcptr->next)
     checkfunc(funcptr, verbosecheck);
 
-  /* Set all primitive units to 1 so valid units will reduce to 1. */
-
-  for(i=0;i<HASHSIZE;i++)
-    for (uptr = utab[i]; uptr; uptr = uptr->next){
-      if (strchr(uptr->value, PRIMITIVECHAR)){
-	free(uptr->value);
-	uptr->value = "1";
-      }
-    }
-
   /* Now check all units for validity */
 
   for(i=0;i<HASHSIZE;i++)
@@ -1700,12 +1880,23 @@ checkunits(int verbosecheck)
       if (verbosecheck)
         printf("doing '%s'\n",uptr->name);
       if (parseunit(&have, uptr->name,0,0) 
-	  || completereduce(&have) || compareunits(&have,&one)){
+	  || completereduce(&have) 
+	  || compareunits(&have,&one, ignore_primitive)){
 	if (isfunction(uptr->name)) 
 	  printf("Unit '%s' hidden by function '%s'\n", uptr->name, uptr->name);
 	else
 	  printf("'%s' defined as '%s' irreducible\n",uptr->name, uptr->value);
+      } else {
+        minusminus = !minusminus;
+        parseunit(&second, uptr->name, 0, 0);
+        completereduce(&second);
+        if (compareunits(&have, &second, ignore_nothing)){
+	  printf("'%s': replace '-' with '+-' for subtraction or '*' to multiply\n", uptr->name);
+	}
+	freeunit(&second);
+        minusminus=!minusminus;
       }
+
       freeunit(&have);
     }
 
@@ -1716,7 +1907,7 @@ checkunits(int verbosecheck)
       if (verbosecheck)
         printf("doing '%s'\n",pptr->name);
       if (parseunit(&have, pptr->name,0,0) 
-	  || completereduce(&have) || compareunits(&have,&one))
+	  || completereduce(&have) || compareunits(&have,&one,ignore_primitive))
 	printf("'%s-' defined as '%s' irreducible\n",pptr->name, pptr->value);
       else { 
 	int plevel;    /* check for bad '/' character in prefix */
@@ -1755,7 +1946,7 @@ void addtolist(struct unittype *have, char *rname, char *name, char *def,
   initializeunit(&want);
   parseunit(&want, name,0,0);
   completereduce(&want);
-  if (!compareunits(have,&want)){
+  if (!compareunits(have,&want,ignore_dimless)){
     if (*count==*listsize){
       *listsize += 100;
       *list = (struct namedef *) 
@@ -1870,14 +2061,18 @@ usage()
         --check-verbose like --check, but lists units as they are checked\n\
                           so you can find units that cause endless loops\n\
     -e, --exponential   exponential format output\n\
-    -f, --file          specify units data file\n\
+    -f, --file          specify units data files (-f '' loads default file)\n\
+    -m, --minus         make - into a subtraction operator (default)\n\
     -o, --output-format specify printf numeric output format\n\
+    -p, --product       make - into a product operator\n\
     -q, --quiet         supress prompting\n\
         --silent        same as --quiet\n\
-    -s, --strict        suppress reciprocal unit conversion\n\
+    -s, --strict        suppress reciprocal unit conversion (e.g. Hz<->s)\n\
+    -t, --terse         print terse conversion factor output\n\
     -v, --verbose       print slightly more verbose output\n\
-    -V, --version       print version number and exit\n", stdout);
-   exit(3);
+    -V, --version       print version number and exit\n\n\
+Report bugs to adrian@cam.cornell.edu.\n\n", stdout);
+   exit(0);
 }
 
 /* Print message about how to get help */
@@ -1892,7 +2087,11 @@ helpmsg()
 void
 printversion()
 {
-  printf("%s version %s %s, units database in %s\n",progname, 
+  printf("GNU Units version %s\n%s, units database in %s\n\
+Copyright (C) 2004 Free Software Foundation, Inc.\n\
+GNU Units comes with ABSOLUTELY NO WARRANTY.\n\
+You may redistribute copies of GNU Units\n\
+under the terms of the GNU General Public License.\n\n", 
 	 VERSION, RVERSTR,UNITSFILE);
 }
 
@@ -2020,7 +2219,48 @@ getuser(char **buffer, int *bufsize, char *query)
 #endif /* READLINE */
 
 
-char *shortoptions = "Vvqechsf:o:";
+char *
+findunitsfile()
+{
+  FILE *testfile;
+  char *file = UNITSFILE;
+  testfile = fopen(file, "rt");
+  if (!testfile) {
+    char *direc, *env;
+    char *filename;
+    char separator[2];
+
+    env = getenv("PATH");
+    if (env) {
+      filename = mymalloc(strlen(env)+strlen(UNITSFILE)+2,"(readunits)");
+
+      if (strchr(env, ';'))	/* MS-DOS */
+	strcpy(separator, ";");
+      else		        /* UNIX */
+	strcpy(separator, ":");
+      direc = strtok(env, separator);
+      while (direc) {
+	strcpy(filename, "");
+	strcat(filename, direc);
+	strcat(filename, "/");
+	strcat(filename, UNITSFILE);
+	testfile = fopen(filename, "rt");
+	if (testfile){
+	  file = dupstr(filename);
+	  break;
+	}
+	direc = strtok(NULL, separator);
+      }
+      free(filename);
+    }
+  }
+  if (testfile)
+    fclose(testfile);
+  return file;
+}
+   
+
+char *shortoptions = "Vvqechstf:o:mp";
 
 struct option longoptions[] = {
   {"version", no_argument, 0, 'V'},
@@ -2035,6 +2275,9 @@ struct option longoptions[] = {
   {"output-format", required_argument, 0, 'o'},
   {"help",no_argument,0,'h'},
   {"strict",no_argument,&strictconvert, 1},
+  {"terse",no_argument, &terseout, 1},
+  {"minus", no_argument, &minusminus, 1},
+  {"product", no_argument, &minusminus, 0},
   {0,0,0,0} };
 
 /* Process the args.  Returns 1 if interactive mode is desired, and 0
@@ -2047,11 +2290,21 @@ processargs(int argc, char **argv, char **from, char **to)
    extern char *optarg;
    extern int optind;
    int optchar, optindex;
+   int ind;
 
    while ( -1 != 
       (optchar = 
          getopt_long(argc, argv,shortoptions,longoptions, &optindex ))) {
       switch (optchar) {
+         case 'm':
+	    minusminus = 1;
+	    break;
+         case 'p':
+	    minusminus = 0;
+	    break;
+         case 't':
+ 	    terseout = 1;
+            break;
          case 'o':
             numformat = optarg;
             break;
@@ -2062,7 +2315,17 @@ processargs(int argc, char **argv, char **from, char **to)
 	    numformat = "%6e";
 	    break;
 	 case 'f':
-	    userfile = optarg;
+ 	    for(ind=0;unitsfiles[ind];ind++); 
+	    if (ind==MAXFILES){
+	      fprintf(stderr, "At most %d -f specifications are allowed\n",
+		      MAXFILES);
+              exit(3);
+	    }
+            if (optarg && *optarg)
+	      unitsfiles[ind] = optarg;
+	    else 
+	      unitsfiles[ind] = findunitsfile();
+	    unitsfiles[ind+1] = 0;
 	    break;
 	 case 'q':
 	    quiet = 1;
@@ -2086,7 +2349,10 @@ processargs(int argc, char **argv, char **from, char **to)
             helpmsg();
       } 
    }
-
+   if (terseout){
+     strictconvert = 1;
+     verbose = 0;
+   }
    if (optind == argc - 2) {
       quiet=1;
       *from = argv[optind];
@@ -2169,6 +2435,7 @@ ishelpquery(char *str, struct unittype *have)
   struct prefixlist *prefix;
   char commandbuf[1000];  /* Hopefully this is enough overkill as no bounds */
   int unitline;           /* checking is performed. */
+  char *file;
   
   str=removepadding(str);
   if (have && !strcmp(str, UNITMATCH)){
@@ -2192,7 +2459,7 @@ Examples:\n\
 \t* 15.24\t\t\t23.889\t\t\t* 1.0660684e+08\n\
 \t/ 0.065\t\t\t\t\t\t/ 9.3802611e-09\n\
 \n\
-The first example shows that 6 inches is about 15 cm (or 1/0.65 cm).\n\
+The first example shows that 6 inches is about 15 cm or (1/0.065) cm.\n\
 The second example shows how to convert 75 degrees Fahrenheit to Celsius.\n\
 \n\
 To quit from units type ^%s.\n\
@@ -2212,18 +2479,23 @@ other units related to myunit. \n\n",
              UNITMATCH);
       return 1;
     }
-    if (function = isfunction(str))
+    if (function = isfunction(str)){
+      file = function->file;
       unitline = function->linenumber;
-    else if (unit = ulookup(str))
+    }
+    else if (unit = ulookup(str)){
       unitline = unit->linenumber;
-    else if ((prefix = plookup(str)) && strlen(str)==prefix->len)
+      file = unit->file;
+    }
+    else if ((prefix = plookup(str)) && strlen(str)==prefix->len){
       unitline = prefix->linenumber;
+      file = prefix->file;
+    }
     else {
       printf("Unknown unit '%s'\n",str);
       return 1;
     }
-    sprintf(commandbuf,"%s +%d %s", pager, unitline, 
-	    userfile?userfile:UNITSFILE);
+    sprintf(commandbuf,"%s +%d %s", pager, unitline, file);
     if (system(commandbuf))
       fprintf(stderr,"%s: unable to invoke pager '%s' to display help\n", 
 	      progname, pager);
@@ -2232,6 +2504,8 @@ other units related to myunit. \n\n",
   return 0;
 }
 
+ 
+
 int
 main(int argc, char **argv)
 {
@@ -2241,22 +2515,50 @@ main(int argc, char **argv)
    int havestrsize=0;   /* Only used if READLINE is undefined */
    int wantstrsize=0;   /* Only used if READLINE is undefined */
    int interactive;
+   int readerr;
+   char **unitfileptr;
+   int unitcount=0, prefixcount=0, funccount=0;   /* for counting units */
 
 #ifdef READLINE
-   rl_completion_entry_function = (Function *)completeunits;
+#  if RL_READLINE_VERSION >= 0x0402 
+      rl_completion_entry_function = (rl_compentry_func_t *)completeunits;
+#  else
+      rl_completion_entry_function = (Function *)completeunits;
+#  endif
    rl_basic_word_break_characters = " \t+-*/()|^";
 #endif
 
-   userfile = getenv("UNITSFILE");
+   unitsfiles[0] = 0;
 
    interactive = processargs(argc, argv, &havestr, &wantstr);
 
+   if (!unitsfiles[0]){
+     unitsfiles[0] = getenv("UNITSFILE");
+     if (!unitsfiles[0])
+       unitsfiles[0] = findunitsfile();
+     unitsfiles[1] = 0;
+   }
+
    mylocale = getenv("LOCALE");
    if (!mylocale)
      mylocale = DEFAULTLOCALE;
 
-   readunits(); /*userfile);*/
+   for(unitfileptr=unitsfiles;*unitfileptr;unitfileptr++){      
+     readerr = readunits(*unitfileptr, stderr, &unitcount, &prefixcount, 
+			 &funccount, 0);
+     if (readerr==E_MEMORY) 
+       exit(3);
+     if (readerr==E_FILE){
+       fprintf(stderr, "%s: unable to open units file '%s'.  ",
+	       progname, *unitfileptr);
+       perror(0);
+       exit(1);
+     }
+   }
 
+   if (!quiet)
+     printf("%d units, %d prefixes, %d nonlinear units\n\n", unitcount, prefixcount,
+     funccount);
    if (unitcheck) {
       checkunits(unitcheck==2 || verbose);
       exit(0);
@@ -2325,19 +2627,12 @@ main(int argc, char **argv)
    return (0);
 }
 
-
 /* NOTES:
 
-mymalloc, growbuffer, and code for reading in a file are the only
-places with print statements that should probably be removed to make a
-library.  How can error reporting in these functions (memory allocation 
-errors) be handled cleanly for a library implementation?  Actually not
-quite try.  The readunits() function and the tryallunits functions can 
-both produce fatal errors as they call realloc to grow buffers.
-
-Consider passing a FILE* into readunit for errors and having all
-functions report errors to that file.  Error reporting can thus be
-suppressed by making it a null pointer.   
+mymalloc, growbuffer and tryallunits are the only places with print
+statements that should (?) be removed to make a library.  How can
+error reporting in these functions (memory allocation errors) be
+handled cleanly for a library implementation?  
 
 method of reporting the definition of a function or table:
   Is it ok as is?  Report inverses?
@@ -2346,26 +2641,10 @@ functions allocated in linked list: could be bad if lots of functions
 
 completion that recognizes built in functions (?)
 
-Additional feature proposals:
-   convert list of numbers from stdin to stdout with the same conversion
-   factor
+Way to report the reduced form of the two operands of a sum when
+they are not conformable.
 
-vs. Jeff's solution:
-
-I haven't had the need; if I did, I'd probably write a simple script; a
-crude example to convert lines in a file 'feet' from ft to m might be:
 */
 
-#if 0
 
-    "sed 's/.*/& ft\"
-    "m/' feet|units -q|sed 's/^[[:space:]]*\* //"
-    "/^[[:space:]]*\//d'"
-
-#endif
-
-/* suggestions:
-    support conversion to a list of units, eg   "ft,in"
-    have units do "spelling advice" if you give a nonconformable conversion
-*/
 
