@@ -1,3 +1,24 @@
+/*
+ *  units, a program for units conversion
+ *  Copyright (C) 1996, 1997, 1999, 2000, 2001 Free Software Foundation, Inc
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+ * 
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+ *     
+ *  This program was written by Adrian Mariano (adrian@cam.cornell.edu)
+ */
+
 #include <math.h>
 #include <errno.h>
 
@@ -53,6 +74,9 @@ double strtod();
 #define E_NOINVERSE 13
 #define E_PARSEMEM 14
 #define E_FUNARGDEF 15
+#define E_FILE 16
+#define E_BADFILE 17
+#define E_MEMORY 18
 
 #define WHITE " \t\n"
 
@@ -97,10 +121,12 @@ struct func {
   char *tableunit;
   struct func *next;
   int linenumber;
+  char *file;                  /* file where defined */ 
 };
 
 extern struct unittype *parameter_value;
 extern char *function_parameter;
+extern int minusminus;
 
 void *mymalloc(int bytes,char *mesg);
 void initializeunit(struct unittype *theunit);
